﻿using System;
using System.Collections.Generic;
using HeuristicLab.Algorithms.PushGP.Expressions.Boolean;
using HeuristicLab.Algorithms.PushGP.Expressions.Code;
using HeuristicLab.Algorithms.PushGP.Expressions.Exec;
using HeuristicLab.Algorithms.PushGP.Expressions.Float;
using HeuristicLab.Algorithms.PushGP.Expressions.Integer;
using HeuristicLab.Algorithms.PushGP.Expressions.Templates;
using HeuristicLab.Algorithms.PushGP.Interpreter;
using HeuristicLab.Algorithms.PushGP.Stack;

namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    public static class ExpressionFactory
    {
        private static Func<IInterpreter, IStack<Expression>> execStackProvider = interpreter => interpreter.ExecStack;
        private static Func<IInterpreter, IStack<Expression>> codeStackProvider = interpreter => interpreter.CodeStack;
        private static Func<IInterpreter, IStack<string>> nameStackProvider = interpreter => interpreter.NameStack;
        private static Func<IInterpreter, IStack<long>> integerStackProvider = interpreter => interpreter.IntegerStack;
        private static Func<IInterpreter, IStack<double>> floatStackProvider = interpreter => interpreter.FloatStack;
        private static Func<IInterpreter, IStack<bool>> booleanStackProvider = interpreter => interpreter.BooleanStack;

        private static IDictionary<OpCode, Expression> expressions = new Dictionary<OpCode, Expression>()
        {
            #region Exec
            { OpCode.ExecPop, new PopExpressionTemplate<Expression>(Symbols.ExecPop, execStackProvider) },
            { OpCode.ExecDuplicate, new DuplicateExpressionTemplate<Expression>(Symbols.ExecDuplicate, execStackProvider) },
            { OpCode.ExecDefine, new ExecDefineExpression() },
            { OpCode.ExecDoXRange, new ExecDoRangeExpression() },
            { OpCode.ExecY, new ExecYExpression() },
            { OpCode.ExecIf, new ExecIfExpression() },
            #endregion Exec

            #region Code
            { OpCode.CodeDuplicate, new DuplicateExpressionTemplate<Expression>(Symbols.CodeDuplicate, codeStackProvider, true) },
            { OpCode.CodeDefine, new CodeDefineExpression() },
            { OpCode.CodeQuote, new CodeQuoteExpression() },
            { OpCode.CodeDo, new CodeDoExpression() },
            { OpCode.CodeIf, new CodeIfExpression() },
            { OpCode.CodeNoop, new CodeNoopExpression() },
            #endregion Code

            #region Name
            { OpCode.NamePop, new PopExpressionTemplate<string>(Symbols.NamePop, nameStackProvider) },
            { OpCode.NameDuplicate, new DuplicateExpressionTemplate<string>(Symbols.NameDuplicate, nameStackProvider) },
            #endregion Name

            #region Boolean
            { OpCode.BooleanPop, new PopExpressionTemplate<bool>(Symbols.BooleanPop, booleanStackProvider) },
            { OpCode.BooleanDuplicate, new DuplicateExpressionTemplate<bool>(Symbols.BooleanDuplicate, booleanStackProvider) },
            { OpCode.BooleanDefine, new DefineExpressionTemplate<bool>(Symbols.BooleanDefine, booleanStackProvider, v => new BooleanPushExpression(v)) },
            { OpCode.BooleanOr, new PushResultExpressionTemplate<bool, bool>(Symbols.BooleanOr, booleanStackProvider, booleanStackProvider, v => v[0] || v[1]) },
            { OpCode.BooleanAnd, new PushResultExpressionTemplate<bool, bool>(Symbols.BooleanAnd, booleanStackProvider, booleanStackProvider, v => v[0] && v[1]) },
            #endregion Boolean

            #region Float
            { OpCode.IntegerPop, new PopExpressionTemplate<long>(Symbols.IntegerPop, integerStackProvider) },
            { OpCode.IntegerDuplicate, new DuplicateExpressionTemplate<long>(Symbols.IntegerDuplicate, integerStackProvider) },
            { OpCode.IntegerDefine, new DefineExpressionTemplate<long>(Symbols.IntegerDefine, integerStackProvider, v => new IntegerPushExpression(v)) },
            { OpCode.IntegerAdd, new PushResultExpressionTemplate<long, long>(Symbols.IntegerAdd, integerStackProvider, integerStackProvider, v => v[0] + v[1]) },
            { OpCode.IntegerSubtract, new PushResultExpressionTemplate<long, long>(Symbols.IntegerSubtract, integerStackProvider, integerStackProvider, v => v[0] - v[1]) },
            { OpCode.IntegerMultiply, new PushResultExpressionTemplate<long, long>(Symbols.IntegerMultiply, integerStackProvider, integerStackProvider, v => v[0] * v[1]) },
            { OpCode.IntegerDivide, new PushResultExpressionTemplate<long, long>(Symbols.IntegerDivide, integerStackProvider, integerStackProvider, v => v[0] / v[1]) },
            { OpCode.IntegerMin, new PushResultExpressionTemplate<long, long>(Symbols.IntegerMin, integerStackProvider, integerStackProvider, v => Math.Min(v[0], v[1])) },
            { OpCode.IntegerMax, new PushResultExpressionTemplate<long, long>(Symbols.IntegerMax, integerStackProvider, integerStackProvider, v => Math.Max(v[0], v[1])) },
            { OpCode.IntegerSmallerThan, new PushResultExpressionTemplate<long, bool>(Symbols.IntegerSmallerThan, integerStackProvider, booleanStackProvider, v => v[0] < v[1]) },
            { OpCode.IntegerGreaterThan, new PushResultExpressionTemplate<long, bool>(Symbols.IntegerGreaterThan, integerStackProvider, booleanStackProvider, v => v[0] > v[1]) },
            { OpCode.IntegerEquals, new PushResultExpressionTemplate<long, bool>(Symbols.IntegerEquals, integerStackProvider, booleanStackProvider, v => v[0] == v[1]) },
            #endregion Float

            #region Float
            { OpCode.FloatPop, new PopExpressionTemplate<double>(Symbols.FloatPop, floatStackProvider) },
            { OpCode.FloatDuplicate, new DuplicateExpressionTemplate<double>(Symbols.FloatDuplicate, floatStackProvider) },
            { OpCode.FloatDefine, new DefineExpressionTemplate<double>(Symbols.FloatDefine, floatStackProvider, v => new FloatPushExpression(v)) },
            { OpCode.FloatAdd, new PushResultExpressionTemplate<double, double>(Symbols.FloatAdd, floatStackProvider, floatStackProvider, v => v[0] + v[1]) },
            { OpCode.FloatSubtract, new PushResultExpressionTemplate<double, double>(Symbols.FloatSubtract, floatStackProvider, floatStackProvider, v => v[0] - v[1]) },
            { OpCode.FloatMultiply, new PushResultExpressionTemplate<double, double>(Symbols.FloatMultiply, floatStackProvider, floatStackProvider, v => v[0] * v[1]) },
            { OpCode.FloatDivide, new PushResultExpressionTemplate<double, double>(Symbols.FloatDivide, floatStackProvider, floatStackProvider, v => v[0] / v[1]) },
            { OpCode.FloatMin, new PushResultExpressionTemplate<double, double>(Symbols.FloatMin, floatStackProvider, floatStackProvider, v => Math.Min(v[0], v[1])) },
            { OpCode.FloatMax, new PushResultExpressionTemplate<double, double>(Symbols.FloatMax, floatStackProvider, floatStackProvider, v => Math.Max(v[0], v[1])) },
            { OpCode.FloatSmallerThan, new PushResultExpressionTemplate<double, bool>(Symbols.FloatSmallerThan, floatStackProvider, booleanStackProvider, v => v[0] < v[1]) },
            { OpCode.FloatGreaterThan, new PushResultExpressionTemplate<double, bool>(Symbols.FloatGreaterThan, floatStackProvider, booleanStackProvider, v => v[0] > v[1]) },
            { OpCode.FloatEquals, new PushResultExpressionTemplate<double, bool>(Symbols.FloatEquals, floatStackProvider, booleanStackProvider, v => v[0] == v[1]) },
            #endregion Float
        };

        public static Expression Create(OpCode opCode)
        {
            Expression expression;
            return expressions.TryGetValue(opCode, out expression)
                ? expression
                : expressions[OpCode.CodeNoop];
        }
    }
}
